//
//  MKThermoCollector.m
//  ThermoViewer-HK
//
//  Created by Matthias Krauß on 24.09.14.
//  Copyright (c) 2014 Matthias Krauß. All rights reserved.
//

#import "MKThermoCollector.h"
#import <HealthKit/HealthKit.h>
#import <CoreBluetooth/CoreBluetooth.h>

#define HTS_SERVICE_UUID @"1809"
#define HTS_MEASURE_CHAR_UUID @"2A1C"

@interface MKThermoCollector () <CBCentralManagerDelegate, CBPeripheralDelegate>

@property (strong) CBCentralManager* centralManager;
@property (strong) CBPeripheral* thermometer;
@property (strong) CBService* thermoService;
@property (strong) CBCharacteristic* thermoMeasureCharacteristic;

@property (strong) HKHealthStore* healthStore;

@property (strong, readwrite) NSString* status;
@property (assign, readwrite) MKThermoCollectorState state;

@end

@implementation MKThermoCollector

- (id) init {
    self = [super init];
    if (!self) return nil;
    self.status = @"Idle";
    self.state = State_off;
    
    return self;
}

- (void) start {
    if (self.state != State_off) return;

    self.state = State_starting;

    if (!self.healthStore) {
        if (![HKHealthStore isHealthDataAvailable]) {
            self.status = @"No HealthKit available";
            return;
        }
        self.healthStore = [HKHealthStore new];
        if (!self.healthStore) {
            self.status = @"HealthKit failed";
        }
    }
    HKQuantityType* tempType = [HKQuantityType quantityTypeForIdentifier:HKQuantityTypeIdentifierBodyTemperature];
    NSSet* writeSet = [NSSet setWithObjects:tempType, nil];
    NSSet* readSet = [NSSet set];
    self.status = @"Launching HealthKit";
    [self.healthStore requestAuthorizationToShareTypes:writeSet readTypes:readSet completion:
    ^(BOOL success, NSError* err){
        [self performSelectorOnMainThread:@selector(hkAuthorizationComplete:) withObject:err waitUntilDone:NO];
    }];
}

- (void) hkAuthorizationComplete:(NSError*)err {
    if (err) {
        self.status = [err localizedDescription];
        self.state = State_failed;
        return;
    }
    HKQuantityType* type = [HKQuantityType quantityTypeForIdentifier:HKQuantityTypeIdentifierBodyTemperature];
    if ([self.healthStore authorizationStatusForType:type] == HKAuthorizationStatusSharingAuthorized) {
        self.status = @"Launching Bluetooth";
    } else {
        self.status = @"Launching without HealthKit";
    }
    if (!self.centralManager) {
        self.centralManager = [[CBCentralManager alloc] initWithDelegate:self queue:nil];
        if (!self.centralManager) {
            self.status = @"Bluetooth failed";
            self.state = State_failed;
        }
    }
}

// CBCentralManagerDelegate
- (void)centralManagerDidUpdateState:(CBCentralManager*)central {
    CBCentralManagerState state = [self.centralManager state];
    switch (state) {
        case CBCentralManagerStatePoweredOff:
            self.state = State_failed;
            self.status = @"Bluetooth is turned off";
            return;
        case CBCentralManagerStateUnsupported:
            self.state = State_failed;
            self.status = @"No Bluetooth";
            return;
        case CBCentralManagerStateUnauthorized:
            self.state = State_failed;
            self.status = @"Bluetooth not allowed";
            return;
        case CBCentralManagerStatePoweredOn:
            [self startScan];
            break;
        default: break; //should be temporary
    }
}

- (void) startScan {
    self.status = @"Scanning";
    self.state = State_disconnected;
    NSArray* services = @[[CBUUID UUIDWithString:HTS_SERVICE_UUID]];
    [self.centralManager scanForPeripheralsWithServices:services options:@{}];
}

// CBCentralManagerDelegate
- (void)centralManager:(CBCentralManager*)central didDiscoverPeripheral:(CBPeripheral*)peripheral advertisementData:(NSDictionary*)advertisementData RSSI:(NSNumber*)RSSI {
    [self.centralManager stopScan];
    self.thermometer = peripheral;
    self.thermometer.delegate = self;
    self.status = @"Connecting";
    [self.centralManager connectPeripheral:peripheral options:@{}];
}

// CBCentralManagerDelegate
- (void) centralManager:(CBCentralManager*)central didFailToConnectPeripheral:(CBPeripheral*)peripheral error:(NSError*)error {
    [self startScan];
}

// CBCentralManagerDelegate
- (void)centralManager:(CBCentralManager*)central didDisconnectPeripheral:(CBPeripheral*)peripheral error:(NSError*)error {
    self.thermometer = nil;
    self.thermoService = nil;
    self.thermoMeasureCharacteristic = nil;
    [self startScan];
}

// CBCentralManagerDelegate
- (void)centralManager:(CBCentralManager*)central didConnectPeripheral:(CBPeripheral*)peripheral {
    self.status = @"Discovering services";
    NSArray* services = @[[CBUUID UUIDWithString:HTS_SERVICE_UUID]];
    [self.thermometer discoverServices:services];
}


//CBPeripheralDelegate
- (void)peripheral:(CBPeripheral*)peripheral didDiscoverServices:(NSError*)error {
    if (error) {
        self.status = error.localizedDescription;
        [self.centralManager cancelPeripheralConnection:peripheral];
    } else {
        if (peripheral.services.count > 0) {
            self.thermoService = peripheral.services[0];
            NSArray* chars = @[[CBUUID UUIDWithString:HTS_MEASURE_CHAR_UUID]];
            self.status = @"Discovering characteristics";
            [self.thermometer discoverCharacteristics:chars forService:self.thermoService];
        } else {
            [self.centralManager cancelPeripheralConnection:peripheral];
        }
    }
}

//CBPeripheralDelegate
- (void) peripheral:(CBPeripheral*)peripheral didDiscoverCharacteristicsForService:(CBService*)service error:(NSError*)error {
    if (error) {
        self.status = error.localizedDescription;
        [self.centralManager cancelPeripheralConnection:peripheral];
    } else {
        if (service.characteristics.count > 0) {
            self.thermoMeasureCharacteristic = service.characteristics[0];
            self.status = @"Subscribing temperature";
            [self.thermometer setNotifyValue:YES forCharacteristic:self.thermoMeasureCharacteristic];
            self.state = State_connected;
        } else {
            [self.centralManager cancelPeripheralConnection:peripheral];
        }
    }
}

- (void)peripheral:(CBPeripheral *)peripheral didUpdateValueForCharacteristic:(CBCharacteristic *)characteristic error:(NSError *)error {
    if (error) {
        self.status = error.localizedDescription;
        [self.centralManager cancelPeripheralConnection:peripheral];
    } else {
        NSData* data = characteristic.value;
        if (data.length >= 5) {
            const uint8_t* buf = data.bytes;
            uint8_t flags = buf[0];
            uint32_t mantissa = buf[1] + (buf[2] << 8) + (buf[3] << 16);
            int8_t exponent = ((int8_t*)buf)[4];
            double value = mantissa * pow(10, exponent);
            NSString* unitString = (flags & 1) ? @"degF" : @"degC";
            self.status = [NSString stringWithFormat:@"Temp: %.1f %@", value, unitString];
            
            HKUnit* unit = [HKUnit unitFromString:unitString];
            HKQuantity* quantity = [HKQuantity quantityWithUnit:unit doubleValue:value];
            HKQuantityType* type = [HKQuantityType quantityTypeForIdentifier:HKQuantityTypeIdentifierBodyTemperature];
            NSDate* now = [NSDate date];
            HKQuantitySample* sample = [HKQuantitySample quantitySampleWithType:type
                                                                       quantity:quantity
                                                                      startDate:now
                                                                        endDate:now
                                                                       metadata:@{}];
            [self.healthStore saveObject:sample withCompletion:^(BOOL success, NSError* err){
            }];
        }
    }
}


@end
